
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "errors.h"
#include "estad.h"
#include "grises.h"
#include "imag.h"
#include "misc.h"
#include "plot.h"
#include "tarea.h"
#include "umbral.h"

int mediaGrVentana (tvent vent)
{
  int ac ;
  int x, y;
  int tx, ty;
  char ** pix;
  tx  = vent->tx ;
  ty  = vent->ty ;
  pix = vent->pix ;
  ac = 0;
  for(x=0;x<tx;x++)
  for(y=0;y<ty;y++){
    ac += (pix[y][x] & 0x0FF) ;
    }
  ac /= tx * ty ;
  return ac ;
}

void histoGr0 (int * gr)
{
  int i;
  for(i=0;i<=255;i++) gr[i]=0;
}

void histoGrVentanaAcum (int * gr, tvent vent)
{
  int x, y;
  int tx, ty;
  char ** pix;
  tx  = vent->tx ;
  ty  = vent->ty ;
  pix = vent->pix ;
  for(x=0;x<tx;x++)
  for(y=0;y<ty;y++){
    gr[pix[y][x] & 0x0FF]++ ;
    }
}

void histoGrVentana (int * gr, tvent vent)
{
  histoGr0 (gr);
  histoGrVentanaAcum (gr, vent);
}

/* comienzo_local */
void histoGrImagAcum (int * gr, timagAcum imagAcum)
{
  int x, y, i;
  int tx, ty;
  int ** pix;
  int ** ni;
  int  n, ndiv2 ;
  tx  = imagAcum->tx ;
  ty  = imagAcum->ty ;
  pix = imagAcum->pix ;
  ni  = imagAcum->ni  ;
  n   = imagAcum->n   ;
  ndiv2 = n/2 ;
  for(i=0;i<=255;i++) gr[i]=0;
  for(x=0;x<tx;x++)
  for(y=0;y<ty;y++)
    if (ni[y][x] == n) {
      gr[ (pix[y][x] + ndiv2) / n ]++ ;
      }
}
/* fin_local */

void histoGrCondVentana (int * gr, tvent vent, int fc)
{
  int gr0[256];
  int i;
  histoGrVentana (gr0, vent);
  for (i=0; i<256; i++)
    gr[i] = 0;
  condensaArrayAcumInt (gr, gr0, 256, fc);
}

void escribeHistoGrSe (int * gr)
{
  int i;
  for(i=0;i<=255;i++)
    printf ("%3d %7d\n", i, gr[i]) ;
}

void escribeHistoGr (char * nf, int * gr)
{
  FILE * fd;
  int i;
  fd = fopen (nf, "w");
  compruebaFd (fd, nf, "grises.escribeHistoGr");
  for(i=0;i<=255;i++)
    fprintf (fd, "%3d %7d\n", i, gr[i]) ;
  fclose (fd);
}

void leeHistoGr (char * nf, int * gr)
{
  FILE * fd;
  int i, j;
  fd = fopen (nf, "r");
  compruebaFd (fd, nf, "grises.leeHistoGr");
  for(i=0;i<=255;i++){
    fscanf (fd, "%3d %7d\n", &j, &(gr[i]) ) ;
    if (j != i)
      printf ("leeHistoGr. leido i: %3d, esperado: %3d\n", j, i);
    if (gr[i] < 0)
      printf ("leeHistoGr. leido gr[i]: %3d, esperado: Positivo o Nulo\n", gr[i]);
    }
  fclose (fd);
}

// nnh: nuevo numero proc hijo gr: array de (256) niv. gris
// nfd: nombre fichero datos nfi: nobre fichero descripcion datos
void presentaHistoGr (int *nnh, int *gr, char *nfd, char *nfi)
{
  escribeHistoGr (nfd, gr);
  escribeComGnuplotSimp (nfi, nfd, "1:2");
  tareaGnuplot (nnh, nfi, "600x400");
}

// vie jul 27 20:52:35 CEST 2007
void posicionMaximosGrises (int *nxb, int *nxn, int * gr)
{
  int mdgr, pmxgr;
  int xb, xn, xb0;
  int k;
  int v[256];
  int p2ng, p2bl;
  int busg;
  mdgr = mediaH (gr, 256);
  // escribeHistoGr("histogr.sal", gr);
  pmxgr = posmaxv (gr, 256);
  // printf ("caracterHistoGris mdgr:%4d, pmxgr:%4d\n", mdgr, pmxgr);  //-
  p2ng = 0; p2bl = 0;
  for(k=0;   k<=mdgr;k++)
    p2ng += (gr[k] * (mdgr-k+1) * (mdgr-k+1)) / 128; // para evitar desbordamiento
  for(k=mdgr+1;k<=255; k++)
    p2bl += (gr[k] * (k- mdgr+1) * (k-mdgr+1)) /128; // para evitar desbordamiento
  // printf ("caracterHistoGris p2ng:%7d, p2bl:%7d\n", p2ng, p2bl);  //-
  if (p2ng > p2bl) {
    // mas blanco
    xb  = pmxgr ;
    // si un punto se ajustase a "e ^ -x*x/k" , calculamos el k para cada punto
    for (k=0;k<xb;k++){
      v[k] = (xb-k)*(xb-k)/(log(gr[xb])-log(gr[k])) ;
      // printf ("%3d %6d\n", k, v[k]);  //-
      }
    busg = min(xb-1,mdgr - 2*(pmxgr - mdgr)/3);
    xn  = posmaxv (v, busg) ;
    }
  else {
    // mas negro
    xn = pmxgr ;
    // si un punto se ajustase a "e ^ -x*x/k" , calculamos el k para cada punto
    for (k=255;k>xn;k--){
      v[k] = (xn-k)*(xn-k)/(log(gr[xn])-log(gr[k])) ;
      // printf ("%3d %6d\n", k, v[k]);  //-
      }
    xb0 = (pmxgr + mdgr) / 2;
    // printf ("xb0=%3d\n",xb0);  //-
    xb  = posmaxv2 (v, 255-xb0, xb0+1) ;
    }
  // printf ("xb=%3d, xn=%3d\n", xb, xn);  //-
  *nxb = xb;
  *nxn = xn;
}

void posicionMaximosGrises3 (int *nxb, int *nxn, int * gr)
{
  int mdgr, pmxgr;
  int xb, xn, xb0;
  int k;
  int v[256];
  mdgr = mediaH (gr, 256);
  pmxgr = posmaxv (gr, 256);
  printf ("caracterHistoGris mdgr:%4d, pmxgr:%4d\n", mdgr, pmxgr);  //-
  if (pmxgr > mdgr) {
    // mas blanco
    xb  = pmxgr ;
    // si un punto se ajustase a "e ^ -x*x/k" , calculamos el k para cada punto
    for (k=0;k<xb;k++){
      v[k] = (xb-k)*(xb-k)/(log(gr[xb])-log(gr[k])) ;
      }
    xn  = posmaxv (v, mdgr - 2*(pmxgr - mdgr)/3) ;
    }
  else {
    // mas negro
    xn = pmxgr ;
    // si un punto se ajustase a "e ^ -x*x/k" , calculamos el k para cada punto
    for (k=255;k>xn;k--){
      v[k] = (xn-k)*(xn-k)/(log(gr[xn])-log(gr[k])) ;
      printf ("%3d %6d\n", k, v[k]);  //-
      }
    xb0 = (pmxgr + mdgr) / 2;
    printf ("xb0=%3d\n",xb0);  //-
    xb  = posmaxv2 (v, 255-xb0, xb0+1) ;
    }
  printf ("xb=%3d, xn=%3d\n", xb, xn);  //-
  *nxb = xb;
  *nxn = xn;
}

void posicionMaximosGrises2 (int *nxb2, int * gr, int xn, int xb)
{
    int k;
    int v[256];
    int xb1, xaux, xb2;
    // printf ("posicionMaximosGrises2 xn: %3d xb: %3d\n", xn, xb);
    for (k=xn; k<=xb; k++) {
      v[k] = (((k - xn) * gr[k]) / 64) * (xb - k);
      }
    xb1 = posmaxv2 (v, xb-xn+1, xn);
    xaux = posminv3 (gr, xb1, xb);
    xb2 = posmaxv2 (gr, xaux-xn+1, xn);
    /*
    printf ("%4d %4d %4d\n", xb1, xaux, xb2);
    */
    *nxb2 = xb2;
}

void desviacionesTipicasGrises (int *ndb, int *ndn, int * gr, int xb, int xn, int hb, int hn)
{
  int i;
  int db, dn;
  if (xb==255)
    db = 1;
  else {
    for(i=xb;(i<=255)&&(gr[i]>(hb/4)||gr[i]==0);i++){}
    if (i<=255)
      db = i - xb;
    else {
      for(i=xb;(i>=0)&&(gr[i]>(hb/3)||gr[i]==0);i--){}
      if (i > xn)
	db = xb -i;
      else
        db = -1;
      }
    }
  if (xn==0)
    dn = 1;
  else {
    for(i=xn;(i>=0)&&(gr[i]>(hn/4)||gr[i]==0);i--){}
    if (i<0)
      dn = -1;
    else
      dn = xn - i;
    }
  *ndb = db;
  *ndn = dn;
}

void desviacionesTipicasGrises2 (int *ndb, int * gr, int xb, int xn, int hb)
{
  int i;
  int db;
  if (xb==255)
    db = 1;
  else {
    for(i=xb;(i>=((2*xb+xn)/3))&&(gr[i]>(hb/3)||gr[i]==0);i--){}
    db = xb - i ;
    }
  *ndb = db;
}

tdcaracHistoGr carHgrNuevo ()
{
  return malloc (sizeof(struct tcaracHistoGr));
}

void caracterizaHistoGris0 (tdcaracHistoGr carHgr, int * gr)
{
                                /** >>> mejorar esta funcion **/
  int ac, acb, acn;
  int db1, db2;
  int xb, xn, hb, hn, db, dn, pmb, pmn;
  posicionMaximosGrises (&xb, &xn, gr);
  // altura maximos
    hn = gr[xn];
    hb = gr[xb];
  // parte blanca y negra   (por mil)
    ac = acum (gr, 256);
    acn = 2 * acum (gr, xn-1) + gr[xn];
    acb = 2 * acum (&gr[xb+1], 255 - xb) + gr[xb];
    pmb = (1000.0 * acb) / ac;
    pmn = (1000.0 * acn) / ac;
  desviacionesTipicasGrises (&db1, &dn, gr, xb, xn, hb, hn);  // por la derecha
  desviacionesTipicasGrises2 (&db2, gr, xb, xn, hb);          // por la izquierda
  if (db1+db2 != 0)
    db = 2 * db1 * db2 / (db1 + db2) ;                      // media armonica
  else
    db = max(db1, db2);
  if ((db >0) && (db < dn) && (pmb < 400)) {   // afinar las condiciones
    int xb2;
    posicionMaximosGrises2 (&xb2, gr, xn, xb);
    xb = xb2;
    hb = gr[xb];
    desviacionesTipicasGrises2 (&db, gr, xb, xn, hb);
    }
  // retorno en parametros tipo puntero
  carHgr->xblanco = xb ;     carHgr->xnegro   = xn ;
  carHgr->alturaB = hb ;     carHgr->alturaN  = hn ;
  carHgr->ppmB    = pmb ;    carHgr->ppmN     = pmn ;
  carHgr->desvB   = db ;     carHgr->desvN    = dn ;
}

//---------- renov vvv

// vie jul 27 20:52:35 CEST 2007
// mar dic 30 10:46:01 CET 2008
void posicionMaximosGrises4 (int *nxb, int *nxn, int * gr)
{
  int mdgr, pmxgr;
  int umbr;
  int xb, xn, xb0;
  int k;
  int v[256];
  int pgn, pgb;
  int busg;
  umbr = umbral (gr, 256);
  mdgr = mediaH (gr, 256);
  // printf ("caracterHistoGris umbr:%7d, mdgr:%7d\n", umbr, mdgr);  //-
  // escribeHistoGr("histogr.sal", gr);
  pmxgr = posmaxv (gr, 256);
  // printf ("caracterHistoGris mdgr:%4d, pmxgr:%4d\n", mdgr, pmxgr);  //-
  pgn = 0; pgb = 0;
  for(k=0; k<=umbr; k++)
    pgn += gr[k] ;
  for(k=umbr+1; k<=255; k++)
    pgb += gr[k] ;
  // printf ("caracterHistoGris pgn:%7d, pgb:%7d\n", pgn, pgb);  //-
  if (pgb > pgn) {
    // mas blanco
    // xb  = pmxgr ;
    xb = posmaxv (&(gr[umbr+1]), 256-umbr-1) + umbr + 1;
    // si un punto se ajustase a "e ^ -x*x/k" , calculamos el k para cada punto
    for (k=0;k<xb;k++){
      v[k] = (xb-k)*(xb-k)/(log(gr[xb])-log(gr[k])) ;
      // printf ("%3d %6d\n", k, v[k]);  //-
      }
    busg = min(xb-1,mdgr - 2*(pmxgr - mdgr)/3);
    xn  = posmaxv (v, busg) ;
    }
  else {
    // mas negro
    // xn = pmxgr ;
    xn = posmaxv (gr, umbr+1);
    // si un punto se ajustase a "e ^ -x*x/k" , calculamos el k para cada punto
    for (k=255;k>xn;k--){
      v[k] = (xn-k)*(xn-k)/(log(gr[xn])-log(gr[k])) ;
      // printf ("%3d %6d\n", k, v[k]);  //-
      }
    xb0 = (pmxgr + mdgr) / 2;
    // printf ("xb0=%3d\n",xb0);  //-
    xb  = posmaxv2 (v, 255-xb0, xb0+1) ;
    }
  // printf ("xb=%3d, xn=%3d\n", xb, xn);  //-
  *nxb = xb;
  *nxn = xn;
}

void caracterizaHistoGris1 (tdcaracHistoGr carHgr, int * gr)
{
                                /** >>> mejorar esta funcion **/
  int ac, acb, acn;
  int db1, db2;
  int xb, xn, hb, hn, db, dn, pmb, pmn;
  posicionMaximosGrises4 (&xb, &xn, gr);
  // altura maximos
    hn = gr[xn];
    hb = gr[xb];
  // parte blanca y negra   (por mil)
    ac = acum (gr, 256);
    acn = 2 * acum (gr, xn-1) + gr[xn];
    acb = 2 * acum (&gr[xb+1], 255 - xb) + gr[xb];
    pmb = (1000.0 * acb) / ac;
    pmn = (1000.0 * acn) / ac;
  desviacionesTipicasGrises (&db1, &dn, gr, xb, xn, hb, hn);  // por la derecha
  desviacionesTipicasGrises2 (&db2, gr, xb, xn, hb);          // por la izquierda
  if (db1+db2 != 0)
    db = 2 * db1 * db2 / (db1 + db2) ;                      // media armonica
  else
    db = max(db1, db2);
  if ((db >0) && (db < dn) && (pmb < 400)) {   // afinar las condiciones
    int xb2;
    posicionMaximosGrises2 (&xb2, gr, xn, xb);
    xb = xb2;
    hb = gr[xb];
    desviacionesTipicasGrises2 (&db, gr, xb, xn, hb);
    }
  // retorno en parametros tipo puntero
  carHgr->xblanco = xb ;     carHgr->xnegro   = xn ;
  carHgr->alturaB = hb ;     carHgr->alturaN  = hn ;
  carHgr->ppmB    = pmb ;    carHgr->ppmN     = pmn ;
  carHgr->desvB   = db ;     carHgr->desvN    = dn ;
}

//---------- renov ^^^

void caracterizaHistoGris(tdcaracHistoGr carHgr, int * gr)
{
//  caracterizaHistoGris0(carHgr, gr);
  caracterizaHistoGris1(carHgr, gr);
/*
  if (carHgr->xblanco < carHgr->xnegro) {
    int gr0[256];
    int i;
    int fc ;
    fc = 2;
    for (i=0; i<256; i++)
      gr0[i] = 0;
    condensaArrayAcumInt (gr0, gr, 256, fc);
    caracterizaHistoGris0(carHgr, gr0);
    carHgr->xblanco  *= fc ;
    carHgr->xnegro   *= fc ;
/ *
    carHgr->alturaB = hb ;
    carHgr->alturaN  = hn ;
    carHgr->ppmB    = pmb ;
    carHgr->ppmN     = pmn ;
* /
    carHgr->desvB   *= fc ;
    carHgr->desvN   *= fc ;
  }
*/
}

tdcaracHistoGr carhisgrDeVentanaCalculado (tvent vent)
{
  int gr[256];
  tdcaracHistoGr carHgr;
  histoGrVentana (gr, vent);
  carHgr = carHgrNuevo ();
  caracterizaHistoGris (carHgr, gr);
  return carHgr;
}

int fondoClaroDeCarhisgrDeVentanaCalculado (tvent vent)
{
  tdcaracHistoGr carHgr;
  carHgr = carhisgrDeVentanaCalculado (vent) ;
  return carHgr -> xblanco;
}

void escribeCaracHistoGr (tdcaracHistoGr carHgr)
{
  printf (" xb: %3d",   carHgr->xblanco) ;
  printf (" db: %3d",   carHgr->desvB) ;
  printf (" pmb: %3d",  carHgr->ppmB) ;
  printf (" hb: %7d\n", carHgr->alturaB) ;
  /**/
  printf (" xn: %3d",   carHgr->xnegro) ;
  printf (" dn: %3d",   carHgr->desvN) ;
  printf (" pmn: %3d",  carHgr->ppmN) ;
  printf (" hn: %7d\n", carHgr->alturaN) ;
}

//  tres tramos rectos, negro, grises, blanco
//  umbrales  n1 y n2
void iniciaConv (int *conv, int n1, int  n2)
{
  int i;
  for(i=0;    i<=n1; i++)   conv[i] = 0;
  for(i=n1+1; i<n2 ; i++)   conv[i] = ((i-n1)*255)/(n2-n1) ;
  for(i=n2;   i<=255; i++)   conv[i] = 255;
}

//  dos tramos planos: negro y blanco, + dos parabolas(semi)
//  centro -> 128,  delta: anchura (semi)parabolas 
//       del -> di(zq) y dd(er) si no caben
void iniciaConv2 (int *conv, int cen, int  del)
{
  int i;
  int di, dd;
  di = del;    dd = del;
  if ((255-cen) < del)
    dd = 255 - cen;
  if (cen < del)
    di = cen;
  for(i=0;        i<=cen-di; i++)
    conv[i] = 0;
  for(i=cen-di+1; i<cen;     i++)
    conv[i] = ((i-cen+di)*(i-cen+di)*128-(di*di)/2)/(di*di);
  conv [cen] = 128;
  for(i=cen+1;    i<cen+dd ; i++)
    conv[i] = 255 - (127*(i-cen-dd)*(i-cen-dd)+(dd*dd)/2)/(dd*dd);
  for(i=cen+dd;   i<=255; i++)
    conv[i] = 255;
}

void iniciaConvCarhis (int * conv, tdcaracHistoGr carHgr)
{
  int xblanco, xnegro;
  int desvB,   desvN;
  xblanco = carHgr->xblanco ;
  xnegro  = carHgr->xnegro ;
  desvB   = carHgr->desvB ;
  desvN   = carHgr->desvN ;
/*
  if ( (desvB+desvN) > 0)
*/
  if ( (desvB > 0) & (desvN > 0) )
    // mayor contraste
    iniciaConv (conv,
                xnegro  + ((xblanco-xnegro)*desvN)/((desvB+desvN)*2),
                xblanco - ((xblanco-xnegro)*desvB)/((desvB+desvN)*2)
               );
  else
    iniciaConv (conv, xnegro, xblanco);
}

tdcaracHistoGr carhisgrDeVentana (tvent vent)
{
  tdcaracHistoGr carHgr;
  tvent v;
  v = vent;
  while (v->carHisGr == 0L) {
    // if (v->venBase == 0L) {
    if (v->venBase == v) {
      carHgr = carhisgrDeVentanaCalculado (v);
      ponerCarhisgrenVentana (v, carHgr);
      return carHgr ;
//      printf ("ocre: error en imag.carhisgrDeVentana, no encontrado carHisGr\n");
//      exit (2);
      }
    v = v->venBase ;
    }
  return v->carHisGr ;
}

int umbralDeVent (tvent vent, int porcNeg)
{
  tdcaracHistoGr carHisGr;
  carHisGr = carhisgrDeVentana (vent);
  return  (carHisGr->xnegro  * porcNeg +
           carHisGr->xblanco * (100 - porcNeg) ) / 100;
}

int umbralDeVent2 (tvent vent, int porcNeg)
{
  tdcaracHistoGr carHisGr;
  tvent nvent;
  ventanaEnVentanaCentrada (&nvent, vent, vent->tx/10, vent->ty/10);
  prepara0 (nvent);
  carHisGr = carhisgrDeVentana (nvent);
  return  (carHisGr->xnegro  * porcNeg +
           carHisGr->xblanco * (100 - porcNeg) ) / 100;
}

void prepara0 (tvent vent)
{
  int gr[256];
  int conv[256];
  struct tcaracHistoGr carHisGr;
  histoGrVentana (gr, vent);
  caracterizaHistoGris ( &carHisGr, gr);
  // if (carHisGr.ppmN > carHisGr.ppmB) {
  //   invierteGrImagen (vent);
  //   histoGrVentana (gr, vent);
  //   caracterizaHistoGris ( &carHisGr, gr);
  //   }
  ponerCarhisgrenVentana (vent, &carHisGr);
  iniciaConvCarhis (conv, &carHisGr);
  ponerConvenVentana (vent, conv);
  // if (d_valori >= 10)
  //   printf ("ocre: segmenta0.prepara0\n") ;
}

