
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "estad.h"
#include "umbral.h"

typedef struct tevas {
  float *va;
  float *vxa;
  float *vx2a;
  float *vx;
  float *vx2;
  float *f;
  } tvas;
typedef tvas *pvas;

pvas nuevoVas (int n)
{
  pvas vas;
  vas = (pvas) malloc (sizeof(tvas));
  vas->va   = (float *) malloc (n * sizeof(float));
  vas->vxa  = (float *) malloc (n * sizeof(float));
  vas->vx2a = (float *) malloc (n * sizeof(float));
  vas->vx   = (float *) malloc (n * sizeof(float));
  vas->vx2  = (float *) malloc (n * sizeof(float));
  vas->f    = (float *) malloc (n * sizeof(float));
  return vas;
}

void calcVas (pvas vas, int* v, int n)
{
  int i;
  vas->vx  [0] = 0;
  vas->vx2 [0] = 0;
  vas->va  [0] = v[0];
  vas->vxa [0] = 0;
  vas->vx2a[0] = 0;
  for (i=1; i<n; i++) {
    vas->vx  [i] = v        [i] * i;
    vas->vx2 [i] = vas->vx  [i] * i;
    vas->va  [i] = vas->va  [i-1] + v[i];
    vas->vxa [i] = vas->vxa [i-1] + vas->vx[i];
    vas->vx2a[i] = vas->vx2a[i-1] + vas->vx2[i];
    }
}

void derivaV (pvas vas, int i0, int i1, int impr)
{
  float vat, vxat; // , vx2at;
  float va0, vxa0; // , vx2a0;
  float *v;
  int i;
  v = vas->f;
  if (i0 > 0) {
    va0   = vas->va  [i0-1];
    vxa0  = vas->vxa [i0-1];
    // vx2a0 = vas->vx2a[i0-1];
    }
  else {
    va0   = 0;
    vxa0  = 0;
    // vx2a0 = 0;
    }
  vat   = vas->va  [i1];
  vxat  = vas->vxa [i1];
  // vx2at = vas->vx2a[i1];
  for (i=i0; i<=i1; i++) {
    if ((vat-vas->va[i] != 0) && (vas->va[i]-va0 != 0))
      v[i] = ((vas->vxa[i] - vxa0) * (vas->vxa[i] - vxa0))/(vas->va[i] - va0) +
             ((vxat - vas->vxa[i]) * (vxat - vas->vxa[i]))/(vat - vas->va[i]) ;
    else
      v[i] = 0;
    if (impr)
      printf ("%4d %11.2f\n", i, v[i]);
    }
}

int umbral (int *v, int n)
{
  pvas vas;
  int umb;
  vas = nuevoVas (n);
  calcVas (vas, v, n);
  derivaV (vas, 0, n-1, 0);
  umb = posrmaxv2 (vas->f, n, 0);
  return umb;
}

tpumbrres umbral2 (int *v, int n)
{
  pvas vas;
  int umb;
  float med, med0, med1;
  tpumbrres pumbrres;
  pumbrres = (tumbrres *) malloc (sizeof (tumbrres));
  vas = nuevoVas (n);
  calcVas (vas, v, n);
  derivaV (vas, 0, n-1, 0);
  pumbrres->umbral = umb = posrmaxv2 (vas->f, n, 0);
  pumbrres->media  = med  = vas->vxa[n-1] / vas->va[n-1];
  pumbrres->media0 = med0 = vas->vxa[umb] / vas->va[umb];
  pumbrres->media1 = med1 = (vas->vxa[n-1] - vas->vxa[umb]) /
                            (vas->va[n-1]  - vas->va[umb]) ;
  pumbrres->desvt  = sqrt ((vas->vx2a[n-1]   / vas->va[n-1]  ) - med  * med );
  pumbrres->desvt0 = sqrt ((vas->vx2a[umb] / vas->va[umb]) - med0 * med0);
  pumbrres->desvt1 = sqrt (((vas->vx2a[n-1] - vas->vx2a[umb]) /
                            (vas->va[n-1]   - vas->va[umb]  )) - med1 * med1);
  return pumbrres;
}
